@@ -18,12 +18,14 @@ from .. import dependencies
 from .. import mesonlib
 from .. import mlog
 import uuid, os, operator
+import typing as T
 
 from ..mesonlib import MesonException
+from ..interpreter import Interpreter
 
 class XCodeBackend(backends.Backend):
-    def __init__(self, build):
-        super().__init__(build)
+    def __init__(self, build: T.Optional[build.Build], interpreter: T.Optional[Interpreter]):
+        super().__init__(build, interpreter)
         self.name = 'xcode'
         self.project_uid = self.environment.coredata.lang_guids['default'].replace('-', '')[:24]
         self.project_conflist = self.gen_id()
@@ -59,7 +61,7 @@ class XCodeBackend(backends.Backend):
         return str(uuid.uuid4()).upper().replace('-', '')[:24]
 
     def get_target_dir(self, target):
-        dirname = os.path.join(target.get_subdir(), self.environment.coredata.get_builtin_option('buildtype'))
+        dirname = os.path.join(target.get_subdir(), self.environment.coredata.get_option(mesonlib.OptionKey('buildtype')))
         os.makedirs(os.path.join(self.environment.get_build_dir(), dirname), exist_ok=True)
         return dirname
 
@@ -74,8 +76,7 @@ class XCodeBackend(backends.Backend):
         if not text.endswith('\n'):
             self.ofile.write('\n')
 
-    def generate(self, interp):
-        self.interpreter = interp
+    def generate(self):
         test_data = self.serialize_tests()[0]
         self.generate_filemap()
         self.generate_buildmap()
@@ -123,7 +124,7 @@ class XCodeBackend(backends.Backend):
     def generate_filemap(self):
         self.filemap = {} # Key is source file relative to src root.
         self.target_filemap = {}
-        for name, t in self.build.targets.items():
+        for name, t in self.build.get_build_targets().items():
             for s in t.sources:
                 if isinstance(s, mesonlib.File):
                     s = os.path.join(s.subdir, s.fname)
@@ -136,7 +137,7 @@ class XCodeBackend(backends.Backend):
 
     def generate_buildmap(self):
         self.buildmap = {}
-        for t in self.build.targets.values():
+        for t in self.build.get_build_targets().values():
             for s in t.sources:
                 s = os.path.join(s.subdir, s.fname)
                 self.buildmap[s] = self.gen_id()
@@ -149,7 +150,7 @@ class XCodeBackend(backends.Backend):
         self.buildstylemap = {'debug': self.gen_id()}
 
     def generate_build_phase_map(self):
-        for tname, t in self.build.targets.items():
+        for tname, t in self.build.get_build_targets().items():
             # generate id for our own target-name
             t.buildphasemap = {}
             t.buildphasemap[tname] = self.gen_id()
@@ -160,7 +161,7 @@ class XCodeBackend(backends.Backend):
 
     def generate_build_configuration_map(self):
         self.buildconfmap = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             bconfs = {'debug': self.gen_id()}
             self.buildconfmap[t] = bconfs
 
@@ -175,18 +176,18 @@ class XCodeBackend(backends.Backend):
 
     def generate_build_configurationlist_map(self):
         self.buildconflistmap = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.buildconflistmap[t] = self.gen_id()
 
     def generate_native_target_map(self):
         self.native_targets = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.native_targets[t] = self.gen_id()
 
     def generate_native_frameworks_map(self):
         self.native_frameworks = {}
         self.native_frameworks_fileref = {}
-        for t in self.build.targets.values():
+        for t in self.build.get_build_targets().values():
             for dep in t.get_external_deps():
                 if isinstance(dep, dependencies.AppleFrameworks):
                     for f in dep.frameworks:
@@ -195,27 +196,27 @@ class XCodeBackend(backends.Backend):
 
     def generate_target_dependency_map(self):
         self.target_dependency_map = {}
-        for tname, t in self.build.targets.items():
+        for tname, t in self.build.get_build_targets().items():
             for target in t.link_targets:
                 self.target_dependency_map[(tname, target.get_basename())] = self.gen_id()
 
     def generate_pbxdep_map(self):
         self.pbx_dep_map = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.pbx_dep_map[t] = self.gen_id()
 
     def generate_containerproxy_map(self):
         self.containerproxy_map = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.containerproxy_map[t] = self.gen_id()
 
     def generate_source_phase_map(self):
         self.source_phase = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.source_phase[t] = self.gen_id()
 
     def generate_pbx_aggregate_target(self):
-        target_dependencies = list(map(lambda t: self.pbx_dep_map[t], self.build.targets))
+        target_dependencies = list(map(lambda t: self.pbx_dep_map[t], self.build.get_build_targets()))
         aggregated_targets = []
         aggregated_targets.append((self.all_id, 'ALL_BUILD', self.all_buildconf_id, [], target_dependencies))
         aggregated_targets.append((self.test_id, 'RUN_TESTS', self.test_buildconf_id, [self.test_command_id], []))
@@ -254,7 +255,7 @@ class XCodeBackend(backends.Backend):
         templ = '%s /* %s */ = { isa = PBXBuildFile; fileRef = %s /* %s */; settings = { COMPILER_FLAGS = "%s"; }; };\n'
         otempl = '%s /* %s */ = { isa = PBXBuildFile; fileRef = %s /* %s */;};\n'
 
-        for t in self.build.targets.values():
+        for t in self.build.get_build_targets().values():
 
             for dep in t.get_external_deps():
                 if isinstance(dep, dependencies.AppleFrameworks):
@@ -263,7 +264,7 @@ class XCodeBackend(backends.Backend):
 
             for s in t.sources:
                 if isinstance(s, mesonlib.File):
-                    s = s.fname
+                    s = os.path.join(s.subdir, s.fname)
 
                 if isinstance(s, str):
                     s = os.path.join(t.subdir, s)
@@ -301,7 +302,7 @@ class XCodeBackend(backends.Backend):
 
     def generate_pbx_container_item_proxy(self):
         self.ofile.write('\n/* Begin PBXContainerItemProxy section */\n')
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.write_line('%s /* PBXContainerItemProxy */ = {' % self.containerproxy_map[t])
             self.indent_level += 1
             self.write_line('isa = PBXContainerItemProxy;')
@@ -315,7 +316,7 @@ class XCodeBackend(backends.Backend):
 
     def generate_pbx_file_reference(self):
         self.ofile.write('\n/* Begin PBXFileReference section */\n')
-        for t in self.build.targets.values():
+        for t in self.build.get_build_targets().values():
             for dep in t.get_external_deps():
                 if isinstance(dep, dependencies.AppleFrameworks):
                     for f in dep.frameworks:
@@ -329,7 +330,7 @@ class XCodeBackend(backends.Backend):
             self.write_line(src_templ % (idval, fullpath, xcodetype, name, path))
         target_templ = '%s /* %s */ = { isa = PBXFileReference; explicitFileType = "%s"; path = %s; refType = %d; sourceTree = BUILT_PRODUCTS_DIR; };\n'
         for tname, idval in self.target_filemap.items():
-            t = self.build.targets[tname]
+            t = self.build.get_build_targets()[tname]
             fname = t.get_filename()
             reftype = 0
             if isinstance(t, build.Executable):
@@ -345,7 +346,7 @@ class XCodeBackend(backends.Backend):
         self.ofile.write('/* End PBXFileReference section */\n')
 
     def generate_pbx_frameworks_buildphase(self):
-        for tname, t in self.build.targets.items():
+        for t in self.build.get_build_targets().values():
             self.ofile.write('\n/* Begin PBXFrameworksBuildPhase section */\n')
             self.write_line('%s /* %s */ = {\n' % (t.buildphasemap['Frameworks'], 'Frameworks'))
             self.indent_level += 1
@@ -367,7 +368,7 @@ class XCodeBackend(backends.Backend):
     def generate_pbx_group(self):
         groupmap = {}
         target_src_map = {}
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             groupmap[t] = self.gen_id()
             target_src_map[t] = self.gen_id()
         self.ofile.write('\n/* Begin PBXGroup section */\n')
@@ -396,7 +397,7 @@ class XCodeBackend(backends.Backend):
         self.write_line('isa = PBXGroup;')
         self.write_line('children = (')
         self.indent_level += 1
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.write_line('%s /* %s */,' % (groupmap[t], t))
         self.indent_level -= 1
         self.write_line(');')
@@ -422,7 +423,7 @@ class XCodeBackend(backends.Backend):
         # write frameworks
         self.indent_level += 1
 
-        for t in self.build.targets.values():
+        for t in self.build.get_build_targets().values():
             for dep in t.get_external_deps():
                 if isinstance(dep, dependencies.AppleFrameworks):
                     for f in dep.frameworks:
@@ -436,7 +437,7 @@ class XCodeBackend(backends.Backend):
         self.write_line('};')
 
         # Targets
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.write_line('%s /* %s */ = {' % (groupmap[t], t))
             self.indent_level += 1
             self.write_line('isa = PBXGroup;')
@@ -454,12 +455,12 @@ class XCodeBackend(backends.Backend):
             self.write_line('isa = PBXGroup;')
             self.write_line('children = (')
             self.indent_level += 1
-            for s in self.build.targets[t].sources:
+            for s in self.build.get_build_targets()[t].sources:
                 s = os.path.join(s.subdir, s.fname)
                 if isinstance(s, str):
                     self.write_line('%s /* %s */,' % (self.filemap[s], s))
-            for o in self.build.targets[t].objects:
-                o = os.path.join(self.build.targets[t].subdir, o)
+            for o in self.build.get_build_targets()[t].objects:
+                o = os.path.join(self.build.get_build_targets()[t].subdir, o)
                 self.write_line('%s /* %s */,' % (self.filemap[o], o))
             self.indent_level -= 1
             self.write_line(');')
@@ -474,7 +475,7 @@ class XCodeBackend(backends.Backend):
         self.write_line('isa = PBXGroup;')
         self.write_line('children = (')
         self.indent_level += 1
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.write_line('%s /* %s */,' % (self.target_filemap[t], t))
         self.indent_level -= 1
         self.write_line(');')
@@ -487,7 +488,7 @@ class XCodeBackend(backends.Backend):
     def generate_pbx_native_target(self):
         self.ofile.write('\n/* Begin PBXNativeTarget section */\n')
         for tname, idval in self.native_targets.items():
-            t = self.build.targets[tname]
+            t = self.build.get_build_targets()[tname]
             self.write_line('%s /* %s */ = {' % (idval, tname))
             self.indent_level += 1
             self.write_line('isa = PBXNativeTarget;')
@@ -503,7 +504,7 @@ class XCodeBackend(backends.Backend):
             self.write_line(');')
             self.write_line('dependencies = (')
             self.indent_level += 1
-            for lt in self.build.targets[tname].link_targets:
+            for lt in self.build.get_build_targets()[tname].link_targets:
                 # NOT DOCUMENTED, may need to make different links
                 # to same target have different targetdependency item.
                 idval = self.pbx_dep_map[lt.get_id()]
@@ -555,7 +556,7 @@ class XCodeBackend(backends.Backend):
         self.indent_level += 1
         self.write_line('%s /* ALL_BUILD */,' % self.all_id)
         self.write_line('%s /* RUN_TESTS */,' % self.test_id)
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             self.write_line('%s /* %s */,' % (self.native_targets[t], t))
         self.indent_level -= 1
         self.write_line(');')
@@ -577,7 +578,7 @@ class XCodeBackend(backends.Backend):
         self.write_line(');')
         self.write_line('runOnlyForDeploymentPostprocessing = 0;')
         self.write_line('shellPath = /bin/sh;')
-        cmd = mesonlib.meson_command + ['test', test_data, '-C', self.environment.get_build_dir()]
+        cmd = mesonlib.get_meson_command() + ['test', test_data, '-C', self.environment.get_build_dir()]
         cmdstr = ' '.join(["'%s'" % i for i in cmd])
         self.write_line('shellScript = "%s";' % cmdstr)
         self.write_line('showEnvVarsInLog = 0;')
@@ -587,15 +588,15 @@ class XCodeBackend(backends.Backend):
 
     def generate_pbx_sources_build_phase(self):
         self.ofile.write('\n/* Begin PBXSourcesBuildPhase section */\n')
-        for name, phase_id in self.source_phase.items():
-            t = self.build.targets[name]
+        for name in self.source_phase.keys():
+            t = self.build.get_build_targets()[name]
             self.write_line('%s /* Sources */ = {' % (t.buildphasemap[name]))
             self.indent_level += 1
             self.write_line('isa = PBXSourcesBuildPhase;')
             self.write_line('buildActionMask = 2147483647;')
             self.write_line('files = (')
             self.indent_level += 1
-            for s in self.build.targets[name].sources:
+            for s in self.build.get_build_targets()[name].sources:
                 s = os.path.join(s.subdir, s.fname)
                 if not self.environment.is_header(s):
                     self.write_line('%s /* %s */,' % (self.buildmap[s], os.path.join(self.environment.get_source_dir(), s)))
@@ -608,7 +609,7 @@ class XCodeBackend(backends.Backend):
 
     def generate_pbx_target_dependency(self):
         targets = []
-        for t in self.build.targets:
+        for t in self.build.get_build_targets():
             idval = self.pbx_dep_map[t] # VERIFY: is this correct?
             targets.append((idval, self.native_targets[t], t, self.containerproxy_map[t]))
 
@@ -634,7 +635,7 @@ class XCodeBackend(backends.Backend):
             self.write_line('isa = XCBuildConfiguration;')
             self.write_line('buildSettings = {')
             self.indent_level += 1
-            self.write_line('ARCHS = "$(ARCHS_STANDARD_32_64_BIT)";')
+            self.write_line('ARCHS = "$(ARCHS_STANDARD_64_BIT)";')
             self.write_line('ONLY_ACTIVE_ARCH = YES;')
             self.write_line('SDKROOT = "macosx";')
             self.write_line('SYMROOT = "%s/build";' % self.environment.get_build_dir())
@@ -702,7 +703,7 @@ class XCodeBackend(backends.Backend):
 
         # Now finally targets.
         langnamemap = {'c': 'C', 'cpp': 'CPLUSPLUS', 'objc': 'OBJC', 'objcpp': 'OBJCPLUSPLUS'}
-        for target_name, target in self.build.targets.items():
+        for target_name, target in self.build.get_build_targets().items():
             for buildtype in self.buildtypes:
                 dep_libs = []
                 links_dylib = False
@@ -733,23 +734,34 @@ class XCodeBackend(backends.Backend):
                 else:
                     product_name = target.get_basename()
                 ldargs += target.link_args
+                linker, stdlib_args = self.determine_linker_and_stdlib_args(target)
+                ldargs += self.build.get_project_link_args(linker, target.subproject, target.for_machine)
+                ldargs += self.build.get_global_link_args(linker, target.for_machine)
+                cargs = []
                 for dep in target.get_external_deps():
+                    cargs += dep.get_compile_args()
                     ldargs += dep.get_link_args()
                 ldstr = ' '.join(ldargs)
                 valid = self.buildconfmap[target_name][buildtype]
                 langargs = {}
-                for lang in self.environment.coredata.compilers:
+                for lang in self.environment.coredata.compilers[target.for_machine]:
                     if lang not in langnamemap:
                         continue
                     # Add compile args added using add_project_arguments()
-                    pargs = self.build.projects_args.get(target.subproject, {}).get(lang, [])
+                    pargs = self.build.projects_args[target.for_machine].get(target.subproject, {}).get(lang, [])
                     # Add compile args added using add_global_arguments()
                     # These override per-project arguments
-                    gargs = self.build.global_args.get(lang, [])
+                    gargs = self.build.global_args[target.for_machine].get(lang, [])
                     targs = target.get_extra_args(lang)
                     args = pargs + gargs + targs
-                    if len(args) > 0:
-                        langargs[langnamemap[lang]] = args
+                    if args:
+                        langname = langnamemap[lang]
+                        compiler = target.compilers.get(lang)
+                        lang_cargs = cargs
+                        if compiler and target.implicit_include_directories:
+                            lang_cargs += self.get_build_dir_include_args(target, compiler)
+                        langargs[langname] = args
+                        langargs[langname] += lang_cargs
                 symroot = os.path.join(self.environment.get_build_dir(), target.subdir)
                 self.write_line('%s /* %s */ = {' % (valid, buildtype))
                 self.indent_level += 1
@@ -783,7 +795,7 @@ class XCodeBackend(backends.Backend):
                         self.write_line('GCC_PREFIX_HEADER = "$(PROJECT_DIR)/%s";' % relative_pch_path)
                 self.write_line('GCC_PREPROCESSOR_DEFINITIONS = "";')
                 self.write_line('GCC_SYMBOLS_PRIVATE_EXTERN = NO;')
-                if len(headerdirs) > 0:
+                if headerdirs:
                     quotedh = ','.join(['"\\"%s\\""' % i for i in headerdirs])
                     self.write_line('HEADER_SEARCH_PATHS=(%s);' % quotedh)
                 self.write_line('INSTALL_PATH = "%s";' % install_path)
@@ -854,7 +866,7 @@ class XCodeBackend(backends.Backend):
         self.indent_level -= 1
         self.write_line('};')
 
-        for target_name in self.build.targets:
+        for target_name in self.build.get_build_targets():
             listid = self.buildconflistmap[target_name]
             self.write_line('%s /* Build configuration list for PBXNativeTarget "%s" */ = {' % (listid, target_name))
             self.indent_level += 1
